@@ -1,5 +1,5 @@
 /*
- * $Id: macbin.h,v 1.2 2001/06/29 14:14:46 rufustfirefly Exp $
+ * $Id: macbin.h,v 1.4 2010-01-27 21:27:53 didg Exp $
  */
 
 #ifndef _MACBIN_H
@@ -10,11 +10,10 @@ struct FHeader;
 
 int bin_open(char *binfile, int flags, struct FHeader *fh, int options);
 int bin_close(int keepflag);
-int bin_read(int fork, char *buffer, int length);
-int bin_write(int fork, char *buffer, int length);
+ssize_t bin_read(int fork, char *buffer, size_t length);
+ssize_t bin_write(int fork, char *buffer, size_t length);
 int bin_header_read(struct FHeader *fh, int revision);
 int bin_header_write(struct FHeader *fh);
 int test_header(void);
-u_short updcrc();
 
 #endif /* _MACBIN_H */
