@@ -1,5 +1,5 @@
 /*
- * $Id: uams_gss.c,v 1.2.2.4.2.2 2009/09/16 09:40:40 franklahm Exp $
+ * $Id: uams_gss.c,v 1.12 2010-03-30 10:25:49 franklahm Exp $
  *
  * Copyright (c) 1990,1993 Regents of The University of Michigan.
  * Copyright (c) 1999 Adrian Sun (asun@u.washington.edu)
@@ -12,7 +12,6 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-#ifndef ATACC
 #include <stdio.h>
 #include <stdlib.h>
 #ifdef HAVE_UNISTD_H
@@ -68,6 +67,10 @@ char *strchr (), *strrchr ();
 #define GSS_C_NT_HOSTBASED_SERVICE gss_nt_service_name
 #endif
 
+#ifdef MIN
+#undef MIN
+#endif
+
 #define MIN(a, b) ((a > b) ? b : a)
 
 static void log_status( char *s, OM_uint32 major_status,
@@ -122,7 +125,10 @@ static void log_ctx_flags( OM_uint32 flags )
 
 static void log_principal(gss_name_t server_name)
 {
-#ifdef DEBUG1
+#if 0
+    if (server_name == GSS_C_NO_NAME)
+        return;
+    /* FIXME: must call gss_canonicalize_name before gss_export_name */
     OM_uint32 major_status = 0, minor_status = 0;
     gss_buffer_desc exported_name;
 
@@ -137,28 +143,31 @@ static void log_principal(gss_name_t server_name)
 static int get_afpd_principal(void *obj, gss_name_t *server_name)
 {
     OM_uint32 major_status = 0, minor_status = 0;
-    char *realm, *fqdn, *service, *principal, *p;
-    int realmlen=0, fqdnlen=0, servicelen=0;
+    char *fqdn, *service, *principal, *p;
+    size_t fqdnlen=0, servicelen=0;
     size_t principal_length;
     gss_buffer_desc s_princ_buffer;
 
-    /* get all the required information from afpd */
-    if (uam_afpserver_option(obj, UAM_OPTION_KRB5REALM, (void*) &realm, &realmlen) < 0)
-        return 1;
+    /* get information from afpd */
     if (uam_afpserver_option(obj, UAM_OPTION_FQDN, (void*) &fqdn, &fqdnlen) < 0)
         return 1;
+    LOG(log_debug, logtype_uams, "get_afpd_principal: fqdn: %s", fqdn);
+
     if (uam_afpserver_option(obj, UAM_OPTION_KRB5SERVICE, (void *)&service, &servicelen) < 0)
         return 1;
+    LOG(log_debug, logtype_uams, "get_afpd_principal: service: %s", service);
 
-    /* we need all the info, log error and return if one's missing */
-    if (!service || !servicelen || !fqdn || !fqdnlen || !realm || !realmlen) {
-        LOG(log_error, logtype_uams,
-            "get_afpd_principal: could not retrieve required information from afpd.");
-        return 1;
+    /* if we don't have all the info, log that and return GSS_C_NO_NAME */
+    if (!service || !servicelen || !fqdn || !fqdnlen) {
+        LOG(log_note, logtype_uams,
+            "get_afpd_principal: could not retrieve information from afpd, using default service principal(s)");
+
+	       *server_name = GSS_C_NO_NAME;
+        return 0;
     }
 
     /* allocate memory to hold the temporary principal string */
-    principal_length = servicelen + 1 + fqdnlen + 1 + realmlen + 1;
+    principal_length = servicelen + 1 + fqdnlen + 1;
     if ( NULL == (principal = (char*) malloc( principal_length)) ) {
         LOG(log_error, logtype_uams,
             "get_afpd_principal: out of memory allocating %u bytes",
@@ -168,24 +177,21 @@ static int get_afpd_principal(void *obj, gss_name_t *server_name)
 
     /*
      * Build the principal string.
-     * Format: 'service/fqdn@realm'
+     * Format: 'service@fqdn'
      */
     strlcpy( principal, service, principal_length);
-    strlcat( principal, "/", principal_length);
+    strlcat( principal, "@", principal_length);
 
     /*
      * The fqdn we get from afpd may contain a port.
      * We need to strip the port from fqdn for principal.
      */
-    p = strchr(fqdn, ':');
-    if (p)
+    if ((p = strchr(fqdn, ':')))
         *p = '\0';
+
     strlcat( principal, fqdn, principal_length);
     if (p)
         *p = ':';
-    strlcat( principal, "@", principal_length);
-    strlcat( principal, realm, principal_length);
-
     /*
      * Import our principal into the gssapi internal representation
      * stored in server_name.
@@ -196,7 +202,7 @@ static int get_afpd_principal(void *obj, gss_name_t *server_name)
     LOG(log_debug, logtype_uams, "get_afpd_principal: importing principal `%s'", principal);
     major_status = gss_import_name( &minor_status,
                                     &s_princ_buffer,
-                                    GSS_C_NO_OID,
+                                    GSS_C_NT_HOSTBASED_SERVICE,
                                     server_name );
 
     /*
@@ -284,7 +290,7 @@ static int wrap_sessionkey(gss_ctx_id_t context, struct session_info *sinfo)
     sesskey_buff.value = sinfo->sessionkey;
     sesskey_buff.length = sinfo->sessionkey_len;
 
-    /* gss_wrap the session key with the default mechanism.
+    /* gss_wrap the session key with the default machanism.
        Require both confidentiality and integrity services */
     gss_wrap (&status, context, 1, GSS_C_QOP_DEFAULT, &sesskey_buff, NULL, &wrap_buff);
 
@@ -316,10 +322,17 @@ static int wrap_sessionkey(gss_ctx_id_t context, struct session_info *sinfo)
 static int acquire_credentials (gss_name_t *server_name, gss_cred_id_t *server_creds)
 {
     OM_uint32 major_status = 0, minor_status = 0;
+    char *envp;
 
-    LOG(log_debug, logtype_uams,
-        "acquire credentials: acquiring credentials (uid = %d, keytab = %s)",
-        (int)geteuid(), getenv( "KRB5_KTNAME") );
+    if ((envp = getenv("KRB5_KTNAME")))
+        LOG(log_debug, logtype_uams,
+            "acquire credentials: acquiring credentials (uid = %d, keytab = %s)",
+            (int)geteuid(), envp);
+    else
+        LOG(log_debug, logtype_uams,
+            "acquire credentials: acquiring credentials (uid = %d) - $KRB5_KTNAME not found in env",
+            (int)geteuid());
+        
     /*
      * Acquire credentials usable for accepting context negotiations.
      * Credentials are for server_name, have an indefinite lifetime,
@@ -437,15 +450,16 @@ static int do_gss_auth(void *obj, char *ibuf, int ticket_len,
     gss_release_cred( &status, &server_creds );
 
 cleanup_vars:
-    gss_release_name( &status, &server_name );
+    if (server_name != GSS_C_NO_NAME)
+        gss_release_name( &status, &server_name );
 
     return ret;
 }
 
 /* -------------------------- */
 static int gss_login(void *obj, struct passwd **uam_pwd,
-                     char *ibuf, int ibuflen,
-                     char *rbuf, int *rbuflen)
+                     char *ibuf, size_t ibuflen,
+                     char *rbuf, size_t *rbuflen)
 {
 
     u_int16_t  temp16;
@@ -462,8 +476,8 @@ static int gss_login(void *obj, struct passwd **uam_pwd,
 }
 
 static int gss_logincont(void *obj, struct passwd **uam_pwd,
-                         char *ibuf, int ibuflen,
-                         char *rbuf, int *rbuflen)
+                         char *ibuf, size_t ibuflen,
+                         char *rbuf, size_t *rbuflen)
 {
     struct passwd *pwd = NULL;
     u_int16_t login_id;
@@ -471,7 +485,7 @@ static int gss_logincont(void *obj, struct passwd **uam_pwd,
     u_int16_t ticket_len;
     char *p;
     int rblen;
-    int userlen;
+    size_t userlen;
     struct session_info *sinfo;
 
     /* Apple's AFP 3.1 documentation specifies that this command
@@ -500,7 +514,7 @@ static int gss_logincont(void *obj, struct passwd **uam_pwd,
 
     rblen = *rbuflen = 0;
 
-    if (ibuflen < 3) {
+    if (ibuflen < 1 +sizeof(login_id)) {
         LOG(log_info, logtype_uams, "uams_gss.c :LoginCont: received incomplete packet");
         return AFPERR_PARAM;
     }
@@ -537,7 +551,7 @@ static int gss_logincont(void *obj, struct passwd **uam_pwd,
 
     if ((ibuf - p + 1) % 2) ibuf++, ibuflen--; /* deal with potential padding */
 
-    LOG(log_info, logtype_uams, "uams_gss.c :LoginCont: client thinks user is %s", p);
+    LOG(log_debug, logtype_uams, "uams_gss.c :LoginCont: client thinks user is %s", p);
 
     /* get the length of the ticket the client sends us */
     memcpy(&ticket_len, ibuf, sizeof(ticket_len));
@@ -559,7 +573,7 @@ static int gss_logincont(void *obj, struct passwd **uam_pwd,
            encoding is the gssapi name in? */
         if((pwd = uam_getname( obj, username, userlen )) == NULL) {
             LOG(log_info, logtype_uams, "uam_getname() failed for %s", username);
-            return AFPERR_PARAM;
+            return AFPERR_NOTAUTH;
         }
         if (uam_checkuser(pwd) < 0) {
             LOG(log_info, logtype_uams, "%s not a valid user", username);
@@ -581,8 +595,8 @@ static int gss_logincont(void *obj, struct passwd **uam_pwd,
  * point is trustworthy as we'll have a signed ticket to parse in logincont.
  */
 static int gss_login_ext(void *obj, char *uname, struct passwd **uam_pwd,
-                         char *ibuf, int ibuflen,
-                         char *rbuf, int *rbuflen)
+                         char *ibuf, size_t ibuflen,
+                         char *rbuf, size_t *rbuflen)
 {
     u_int16_t  temp16;
 
@@ -622,4 +636,3 @@ UAM_MODULE_EXPORT struct uam_export uams_gss = {
     UAM_MODULE_VERSION,
     uam_setup, uam_cleanup
 };
-#endif
