@@ -1,5 +1,5 @@
 /*
- * $Id: ad_size.c,v 1.5.6.1 2004/02/20 19:57:15 didg Exp $
+ * $Id: ad_size.c,v 1.8 2010-02-26 14:13:16 didg Exp $
  *
  * Copyright (c) 1997 Adrian Sun (asun@zoology.washington.edu)
  * All rights reserved. See COPYRIGHT.
@@ -11,6 +11,7 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
+#include <string.h>
 #include <atalk/logger.h>
 #include <atalk/adouble.h>
 
@@ -18,8 +19,11 @@ off_t ad_size(const struct adouble *ad, const u_int32_t eid)
 {
   if (eid == ADEID_DFORK) {
     struct stat st;
-    
-    if (fstat(ad_dfileno(ad), &st) < 0)
+
+    if (ad->ad_data_fork.adf_syml) 
+        return strlen(ad->ad_data_fork.adf_syml);
+
+    if (fstat(ad_data_fileno(ad), &st) < 0)
       return 0;
     return st.st_size;
   }  
